/*
 * An opener for a "host" file provided via the C library.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include <fsclient/client.h>
#include <resource/resource_server.h>

#include "host_directory_accessor.h"
#include "host_file_accessor.h"
#include "host_file_opener.h"



HostFileOpener::~HostFileOpener()
{
}

bool HostFileOpener::accessing_directory(flags_t flags, fileid_t fileid)
{
    (void) flags;

    const char *path = _get_path(fileid);

    if (path == NULL)
        return false;

    struct stat st;

    if (::stat(path, &st))
        return false;

    return (st.st_mode & S_IFDIR) ? true : false;
}

bool HostFileOpener::accessing_file(flags_t flags, fileid_t fileid)
{
    (void) flags;
    struct stat st;

    const char *path = _get_path(fileid);

    if (path == NULL)
        return false;

    if (::stat(path, &st))
        return false;

    return (st.st_mode & S_IFREG) ? true : false;
}

/* Test if a directory is empty. */

bool HostFileOpener::directory_is_empty(fileid_t fileid)
{
    const char *path = _get_path(fileid);

    if (path == NULL)
        return false;

    DIR *dir = opendir(path);
    struct dirent *entry;

    if (dir == NULL)
        return false;

    while ((entry = readdir(dir)) != NULL)
    {
        if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, ".."))
        {
            closedir(dir);
            return false;
        }
    }

    closedir(dir);
    return true;
}

/* Return a file identifier for the given 'path'. */

long HostFileOpener::get_fileid(const char *path, flags_t flags, fileid_t *fileid)
{
    (void) flags;

    /* Test for a valid file. */

    FILE *fp = fopen(path, "r");

    if (fp == NULL)
        return -L4_ENOENT;

    fclose(fp);

    *fileid = _get_fileid(path, true);
    return L4_EOK;
}

fileid_t HostFileOpener::_get_fileid(const char *path, bool create)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* The inode number would be a good choice, but the L4Re read-only file
       implementation just provides the value of a counter!

       See: pkg/l4re-core/l4re_vfs/include/impl/ro_file_impl.h */

    std::string s(path);

    HostFileIdentifiers::iterator it = _fileids.find(s);

    if (it != _fileids.end())
        return it->second;

    if (!create)
        return FILEID_INVALID;

    fileid_t fileid = _fileids.size();

    _fileids[s] = fileid;
    _paths[fileid] = s;

    return fileid;
}

const char *HostFileOpener::_get_path(fileid_t fileid)
{
    std::lock_guard<std::mutex> guard(_lock);

    HostFilePaths::iterator it = _paths.find(fileid);

    if (it != _paths.end())
        return it->second.c_str();
    else
        return NULL;
}

/* Return a new accessor for 'fileid'. */

long HostFileOpener::make_accessor(flags_t flags, fileid_t fileid,
                                   Accessor **accessor)
{
    // NOTE: Not testing for create or write flags.

    (void) flags;

    const char *path = _get_path(fileid);

    if (path == NULL)
        return -L4_ENOENT;

    FILE *fp = fopen(path, "r");

    if (fp == NULL)
        return -L4_ENOENT;

    *accessor = new HostFileAccessor(fp, fileid);
    return L4_EOK;
}

/* Return a directory accessor for 'fileid'. */

long HostFileOpener::make_directory_accessor(flags_t flags, fileid_t fileid,
                                             DirectoryAccessor **accessor)
{
    (void) flags;

    const char *path = _get_path(fileid);

    if (path == NULL)
        return -L4_ENOENT;

    *accessor = new HostDirectoryAccessor(path);
    return L4_EOK;
}



/* Make a new directory. */

long HostFileOpener::make_directory_object(const char *path, sys_mode_t mode)
{
    int err = mkdir(path, systypes_from_sys_mode(mode));

    // NOTE: Other return codes may need converting.

    switch (err)
    {
        case 0: break;
        case EACCES: return -L4_EACCESS;
        case EEXIST: return -L4_EEXIST;
        case ENOENT: return -L4_ENOENT;
        default: return -L4_EIO;
    }

    fileid_t fileid = _get_fileid(path, true);

    _fileids[path] = fileid;
    _paths[fileid] = path;

    return L4_EOK;
}

/* Remove a filesystem object. */

long HostFileOpener::remove_object(fileid_t fileid)
{
    (void) fileid;
    return L4_EOK;
}

/* Rename a filesystem object, placing source inside the parent of target. */

long HostFileOpener::rename_object(const char *source, const char *target)
{
    /* NOTE: Should propagate a more meaningful error. */

    if (rename(source, target))
        return -L4_EIO;

    return L4_EOK;
}

/* Populate a memory region with statistics metadata for a filesystem object. */

long HostFileOpener::stat_object(const char *path, void *base, offset_t size)
{
    struct stat *st = (struct stat *) base;

    if (sizeof(struct stat) > size)
        return -L4_ENOMEM;

    return ::stat(path, st);
}

/* Unlink a filesystem object. */

long HostFileOpener::unlink_object(fileid_t parent_fileid, fileid_t fileid)
{
    /* Ignore the parent and always use the path. */

    (void) parent_fileid;

    const char *path = _get_path(fileid);

    if (path == NULL)
        return -L4_ENOENT;

    // NOTE: Return code may need converting.

    int err = unlink(path);

    if (err)
        return err;

    /* Remove the file identifier mapping. */

    std::string s(path);

    HostFileIdentifiers::iterator it = _fileids.find(s);

    if (it != _fileids.end())
        _fileids.erase(it);

    HostFilePaths::iterator pit = _paths.find(fileid);

    if (pit != _paths.end())
        _paths.erase(pit);

    return L4_EOK;
}

// vim: tabstop=4 expandtab shiftwidth=4
